@@ -2,11 +2,34 @@
 # this repository contains the full copyright notices and license terms.
 import datetime
 
-from trytond.model import ModelView, fields
-from trytond.pyson import Eval
-from trytond.pool import Pool, PoolMeta
-from trytond.transaction import Transaction
+from trytond.model import ModelSQL, ModelView, fields
+from trytond.modules.company.model import CompanyValueMixin
+from trytond.modules.currency.fields import Monetary
 from trytond.modules.product import price_digits
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Eval
+from trytond.transaction import Transaction
+
+
+class Configuration(metaclass=PoolMeta):
+    __name__ = 'product.configuration'
+
+    default_lead_time = fields.MultiValue(
+        fields.TimeDelta(
+            "Default Lead Time",
+            help="The time from confirming the sales order to sending the "
+            "products.\n"
+            "Used for products without a lead time."))
+
+    @classmethod
+    def default_default_lead_time(cls, **pattern):
+        return datetime.timedelta(0)
+
+
+class DefaultLeadTime(ModelSQL, CompanyValueMixin):
+    "Product Default Lead Time"
+    __name__ = 'product.configuration.default_lead_time'
+    default_lead_time = fields.TimeDelta("Default Lead Time")
 
 
 class Template(metaclass=PoolMeta):
@@ -20,18 +43,21 @@ class Template(metaclass=PoolMeta):
             },
         domain=[
             ('category', '=', Eval('default_uom_category')),
-            ],
-        depends=['salable', 'default_uom_category'])
+            ])
     lead_time = fields.TimeDelta(
         "Lead Time",
         states={
             'invisible': ~Eval('salable', False),
             },
-        depends=['salable'])
+        help="The time from confirming the sales order to sending the "
+        "products.\n"
+        "If empty the default lead time from the configuration is used.")
 
     @classmethod
     def __register__(cls, module_name):
-        cursor = Transaction().connection.cursor()
+        transaction = Transaction()
+        cursor = transaction.connection.cursor()
+        update = transaction.connection.cursor()
         table = cls.__table_handler__(module_name)
         sql_table = cls.__table__()
 
@@ -41,20 +67,16 @@ class Template(metaclass=PoolMeta):
         if table.column_exist('delivery_time'):
             cursor.execute(*sql_table.select(
                     sql_table.id, sql_table.delivery_time))
-            for id_, delivery_time in cursor.fetchall():
+            for id_, delivery_time in cursor:
                 if delivery_time is None:
                     continue
                 lead_time = datetime.timedelta(days=delivery_time)
-                cursor.execute(*sql_table.update(
+                update.execute(*sql_table.update(
                         [sql_table.lead_time],
                         [lead_time],
                         where=sql_table.id == id_))
             table.drop_column('delivery_time')
 
-    @staticmethod
-    def default_lead_time():
-        return datetime.timedelta(0)
-
     @fields.depends('default_uom', 'sale_uom', 'salable')
     def on_change_default_uom(self):
         try:
@@ -79,7 +101,7 @@ class Template(metaclass=PoolMeta):
 class Product(metaclass=PoolMeta):
     __name__ = 'product.product'
 
-    sale_price_uom = fields.Function(fields.Numeric(
+    sale_price_uom = fields.Function(Monetary(
             "Sale Price", digits=price_digits), 'get_sale_price_uom')
 
     @classmethod
@@ -120,23 +142,35 @@ class Product(metaclass=PoolMeta):
         user = User(Transaction().user)
 
         for product in products:
-            prices[product.id] = product._get_sale_unit_price(
-                quantity=quantity)
-            if uom and product.default_uom.category == uom.category:
-                prices[product.id] = Uom.compute_price(
-                    product.default_uom, prices[product.id], uom)
-            else:
-                prices[product.id] = Uom.compute_price(
-                    product.default_uom, prices[product.id], product.sale_uom)
-            if currency and user.company:
+            unit_price = product._get_sale_unit_price(quantity=quantity)
+            if unit_price is not None:
+                if uom and product.default_uom.category == uom.category:
+                    unit_price = Uom.compute_price(
+                        product.default_uom, unit_price, uom)
+                else:
+                    unit_price = Uom.compute_price(
+                        product.default_uom, unit_price, product.sale_uom)
+            if currency and user.company and unit_price is not None:
                 if user.company.currency != currency:
                     date = Transaction().context.get('sale_date') or today
                     with Transaction().set_context(date=date):
-                        prices[product.id] = Currency.compute(
-                            user.company.currency, prices[product.id],
+                        unit_price = Currency.compute(
+                            user.company.currency, unit_price,
                             currency, round=False)
+            prices[product.id] = unit_price
         return prices
 
+    @property
+    def lead_time_used(self):
+        pool = Pool()
+        Configuration = pool.get('product.configuration')
+        if self.lead_time is None:
+            with Transaction().set_context(self._context):
+                config = Configuration(1)
+                return config.get_multivalue('default_lead_time')
+        else:
+            return self.lead_time
+
     def compute_shipping_date(self, date=None):
         '''
         Compute the shipping date at the given date
@@ -144,10 +178,13 @@ class Product(metaclass=PoolMeta):
         Date = Pool().get('ir.date')
 
         if not date:
-            date = Date.today()
-        if self.lead_time is None:
+            with Transaction().set_context(context=self._context):
+                date = Date.today()
+
+        lead_time = self.lead_time_used
+        if lead_time is None:
             return datetime.date.max
-        return date + self.lead_time
+        return date + lead_time
 
 
 class SaleContext(ModelView):
@@ -159,7 +196,12 @@ class SaleContext(ModelView):
         domain=[('type', '=', 'warehouse')])
     company = fields.Many2One('company.company', "Company")
     currency = fields.Many2One('currency.currency', "Currency")
-    customer = fields.Many2One('party.party', "Customer")
+    customer = fields.Many2One(
+        'party.party', "Customer",
+        context={
+            'company': Eval('company', -1),
+            },
+        depends={'company'})
     sale_date = fields.Date("Sale Date")
     quantity = fields.Float("Quantity")
 
