@@ -3,32 +3,31 @@
 import datetime
 from collections import defaultdict
 from decimal import Decimal
-from itertools import groupby, chain
 from functools import partial
+from itertools import chain, groupby
 
 from trytond.i18n import gettext
-from trytond.model import Workflow, Model, ModelView, ModelSQL, fields, \
-    sequence_ordered, Unique
-from trytond.model.exceptions import AccessError
-from trytond.modules.company import CompanyReport
-from trytond.wizard import Wizard, StateAction, StateView, StateTransition, \
-    Button
-from trytond.pyson import If, Eval, Bool, PYSONEncoder
-from trytond.transaction import Transaction
-from trytond.tools import sortable_values, firstline
-from trytond.pool import Pool
-
 from trytond.ir.attachment import AttachmentCopyMixin
 from trytond.ir.note import NoteCopyMixin
+from trytond.model import (
+    Model, ModelSQL, ModelView, Unique, Workflow, fields, sequence_ordered)
+from trytond.model.exceptions import AccessError
 from trytond.modules.account.tax import TaxableMixin
 from trytond.modules.account_product.exceptions import AccountError
+from trytond.modules.company import CompanyReport
 from trytond.modules.company.model import (
-    employee_field, set_employee, reset_employee)
+    employee_field, reset_employee, set_employee)
 from trytond.modules.currency.fields import Monetary
 from trytond.modules.product import price_digits, round_price
+from trytond.pool import Pool
+from trytond.pyson import Bool, Eval, If, PYSONEncoder
+from trytond.tools import firstline, sortable_values
+from trytond.transaction import Transaction
+from trytond.wizard import (
+    Button, StateAction, StateTransition, StateView, Wizard)
 
 from .exceptions import (
-    SaleValidationError, SaleQuotationError, PartyLocationError)
+    PartyLocationError, SaleQuotationError, SaleValidationError)
 
 
 def get_shipments_returns(model_name):
@@ -76,28 +75,24 @@ class Sale(
                 | Eval('party', True)
                 | Eval('invoice_party', True)
                 | Eval('shipment_party', True)),
-            },
-        depends=['state'])
+            })
     number = fields.Char('Number', readonly=True, select=True)
     reference = fields.Char('Reference', select=True)
     description = fields.Char('Description',
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     sale_date = fields.Date('Sale Date',
         states={
             'readonly': ~Eval('state').in_(['draft', 'quotation']),
             'required': ~Eval('state').in_(
                 ['draft', 'quotation', 'cancelled']),
-            },
-        depends=['state'])
+            })
     payment_term = fields.Many2One('account.invoice.payment_term',
         'Payment Term',
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     party = fields.Many2One('party.party', 'Party', required=True, select=True,
         states={
             'readonly': ((Eval('state') != 'draft')
@@ -106,7 +101,7 @@ class Sale(
         context={
             'company': Eval('company', -1),
             },
-        depends=['state', 'company'])
+        depends={'company'})
     party_lang = fields.Function(fields.Char('Party Language'),
         'on_change_with_party_lang')
     contact = fields.Many2One(
@@ -117,7 +112,7 @@ class Sale(
         search_context={
             'related_party': Eval('party'),
             },
-        depends=['party', 'company'])
+        depends={'company'})
     invoice_party = fields.Many2One('party.party', "Invoice Party",
         states={
             'readonly': ((Eval('state') != 'draft')
@@ -129,7 +124,7 @@ class Sale(
         search_context={
             'related_party': Eval('party'),
             },
-        depends=['state', 'party', 'company'])
+        depends={'company'})
     invoice_address = fields.Many2One('party.address', 'Invoice Address',
         domain=[
             ('party', '=', If(Bool(Eval('invoice_party')),
@@ -139,7 +134,7 @@ class Sale(
             'readonly': Eval('state') != 'draft',
             'required': ~Eval('state').in_(
                 ['draft', 'quotation', 'cancelled']),
-            }, depends=['party', 'invoice_party', 'state'])
+            })
     shipment_party = fields.Many2One('party.party', 'Shipment Party',
         states={
             'readonly': (Eval('state') != 'draft'),
@@ -150,7 +145,7 @@ class Sale(
         search_context={
             'related_party': Eval('party'),
             },
-        depends=['state', 'party', 'company'])
+        depends={'company'})
     shipment_address = fields.Many2One('party.address', 'Shipment Address',
         domain=[
             ('party', '=', If(Bool(Eval('shipment_party')),
@@ -158,23 +153,20 @@ class Sale(
             ],
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['party', 'shipment_party', 'state'])
+            })
     warehouse = fields.Many2One('stock.location', 'Warehouse',
         domain=[('type', '=', 'warehouse')], states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     currency = fields.Many2One('currency.currency', 'Currency', required=True,
         states={
             'readonly': ((Eval('state') != 'draft')
                 | (Eval('lines', [0]) & Eval('currency', 0))),
-            },
-        depends=['state'])
+            })
     lines = fields.One2Many('sale.line', 'sale', 'Lines', states={
             'readonly': Eval('state') != 'draft',
             },
-        depends=['party', 'state'])
+        depends={'party'})
     comment = fields.Text('Comment')
     untaxed_amount = fields.Function(Monetary(
             "Untaxed", digits='currency', currency='currency'), 'get_amount')
@@ -195,15 +187,14 @@ class Sale(
             ],
         'Invoice Method', required=True, states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     invoice_method_string = invoice_method.translated('invoice_method')
     invoice_state = fields.Selection([
             ('none', 'None'),
             ('waiting', 'Waiting'),
             ('paid', 'Paid'),
             ('exception', 'Exception'),
-            ], 'Invoice State', readonly=True, required=True)
+            ], 'Invoice State', readonly=True, required=True, sort=False)
     invoices = fields.Function(fields.Many2Many(
             'account.invoice', None, None, "Invoices"),
         'get_invoices', searcher='search_invoices')
@@ -219,27 +210,27 @@ class Sale(
             ], 'Shipment Method', required=True,
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     shipment_method_string = shipment_method.translated('shipment_method')
     shipment_state = fields.Selection([
             ('none', 'None'),
             ('waiting', 'Waiting'),
             ('sent', 'Sent'),
             ('exception', 'Exception'),
-            ], 'Shipment State', readonly=True, required=True)
+            ], "Shipment State", readonly=True, required=True, sort=False)
     shipments = fields.Function(fields.Many2Many(
             'stock.shipment.out', None, None, "Shipments"),
         'get_shipments', searcher='search_shipments')
     shipment_returns = fields.Function(fields.Many2Many(
             'stock.shipment.out.return', None, None, "Shipment Returns"),
         'get_shipment_returns', searcher='search_shipment_returns')
-    moves = fields.One2Many('stock.move', 'sale', 'Moves', readonly=True)
+    moves = fields.Function(
+        fields.Many2Many('stock.move', None, None, "Moves"),
+        'get_moves', searcher='search_moves')
     origin = fields.Reference('Origin', selection='get_origin', select=True,
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     shipping_date = fields.Date(
         "Shipping Date",
         domain=[If(Bool(Eval('sale_date')) & Bool(Eval('shipping_date')),
@@ -249,7 +240,6 @@ class Sale(
         states={
             'readonly': Eval('state').in_(['processing', 'done', 'cancelled']),
             },
-        depends=['state', 'sale_date'],
         help="When the shipping of goods should start.")
 
     quoted_by = employee_field(
@@ -265,7 +255,7 @@ class Sale(
             ('processing', "Processing"),
             ('done', "Done"),
             ('cancelled', "Cancelled"),
-            ], "State", readonly=True, required=True)
+            ], "State", readonly=True, required=True, sort=False)
     state_string = state.translated('state')
 
     @classmethod
@@ -390,10 +380,13 @@ class Sale(
     def default_state():
         return 'draft'
 
-    @staticmethod
-    def default_currency():
-        Company = Pool().get('company.company')
-        company = Transaction().context.get('company')
+    @classmethod
+    def default_currency(cls, **pattern):
+        pool = Pool()
+        Company = pool.get('company.company')
+        company = pattern.get('company')
+        if not company:
+            company = cls.default_company()
         if company:
             return Company(company).currency.id
 
@@ -417,13 +410,19 @@ class Sale(
     def default_shipment_state():
         return 'none'
 
-    @fields.depends('party', 'invoice_party', 'shipment_party', 'payment_term')
+    @fields.depends(
+        'company', 'party', 'invoice_party', 'shipment_party', 'payment_term',
+        'lines')
     def on_change_party(self):
         if not self.invoice_party:
             self.invoice_address = None
         if not self.shipment_party:
             self.shipment_address = None
-        self.payment_term = self.default_payment_term()
+        self.payment_term = self.default_payment_term(
+            company=self.company.id if self.company else None)
+        if not self.lines:
+            self.currency = self.default_currency(
+                company=self.company.id if self.company else None)
         if self.party:
             if not self.invoice_party:
                 self.invoice_address = self.party.address_get(type='invoice')
@@ -439,6 +438,9 @@ class Sale(
                 self.invoice_method = self.party.sale_invoice_method
             else:
                 self.invoice_method = self.default_invoice_method()
+            if not self.lines:
+                if self.party.customer_currency:
+                    self.currency = self.party.customer_currency
 
     @fields.depends('party', 'invoice_party')
     def on_change_invoice_party(self):
@@ -594,6 +596,13 @@ class Sale(
                 return 'waiting'
         return 'none'
 
+    def get_moves(self, name):
+        return [m.id for l in self.lines for m in l.moves]
+
+    @classmethod
+    def search_moves(cls, name, clause):
+        return [('lines.' + clause[0],) + tuple(clause[1:])]
+
     @classmethod
     def _get_origin(cls):
         'Return list of Model names for origin Reference'
@@ -614,29 +623,33 @@ class Sale(
             return ''
 
     @classmethod
-    def validate(cls, sales):
-        super(Sale, cls).validate(sales)
-        for sale in sales:
-            sale.check_method()
+    def validate_fields(cls, sales, field_names):
+        super().validate_fields(sales, field_names)
+        cls.check_method(sales, field_names)
 
-    def check_method(self):
+    @classmethod
+    def check_method(cls, sales, field_names=None):
         '''
         Check the methods.
         '''
-        if (self.invoice_method == 'shipment'
-                and self.shipment_method in ('invoice', 'manual')):
-            raise SaleValidationError(
-                gettext('sale.msg_sale_invalid_method',
-                    invoice_method=self.invoice_method_string,
-                    shipment_method=self.shipment_method_string,
-                    sale=self.rec_name))
-        if (self.shipment_method == 'invoice'
-                and self.invoice_method in ('shipment', 'manual')):
-            raise SaleValidationError(
-                gettext('sale.msg_sale_invalid_method',
-                    invoice_method=self.invoice_method_string,
-                    shipment_method=self.shipment_method_string,
-                    sale=self.rec_name))
+        if field_names and not (field_names & {
+                    'invoice_method', 'shipment_method'}):
+            return
+        for sale in sales:
+            if (sale.invoice_method == 'shipment'
+                    and sale.shipment_method in {'invoice', 'manual'}):
+                raise SaleValidationError(
+                    gettext('sale.msg_sale_invalid_method',
+                        invoice_method=sale.invoice_method_string,
+                        shipment_method=sale.shipment_method_string,
+                        sale=sale.rec_name))
+            if (sale.shipment_method == 'invoice'
+                    and sale.invoice_method in {'shipment', 'manual'}):
+                raise SaleValidationError(
+                    gettext('sale.msg_sale_invalid_method',
+                        invoice_method=sale.invoice_method_string,
+                        shipment_method=sale.shipment_method_string,
+                        sale=sale.rec_name))
 
     @property
     def full_number(self):
@@ -704,7 +717,6 @@ class Sale(
         default.setdefault('number', None)
         default.setdefault('invoice_state', 'none')
         default.setdefault('invoices_ignored', None)
-        default.setdefault('moves', None)
         default.setdefault('shipment_state', 'none')
         default.setdefault('sale_date', None)
         default.setdefault('quoted_by')
@@ -753,9 +765,12 @@ class Sale(
     @classmethod
     def set_sale_date(cls, sales):
         Date = Pool().get('ir.date')
-        cls.write([s for s in sales if not s.sale_date], {
-                'sale_date': Date.today(),
-                })
+        for company, c_sales in groupby(sales, key=lambda s: s.company):
+            with Transaction().set_context(company=company.id):
+                today = Date.today()
+            cls.write([s for s in c_sales if not s.sale_date], {
+                    'sale_date': today,
+                    })
 
     @classmethod
     def store_cache(cls, sales):
@@ -798,6 +813,10 @@ class Sale(
         if getattr(invoice, 'lines', None):
             invoice_lines = list(invoice.lines) + invoice_lines
         invoice.lines = invoice_lines
+        invoice.save()
+
+        invoice.update_taxes()
+        self.copy_resources_to(invoice)
         return invoice
 
     def _group_shipment_key(self, moves, move):
@@ -862,6 +881,8 @@ class Sale(
             shipment = self._get_shipment_sale(Shipment, key)
             shipment.moves = (list(getattr(shipment, 'moves', []))
                 + [x[1] for x in grouped_moves])
+            shipment.save()
+            self.copy_resources_to(shipment)
             shipments.append(shipment)
         return shipments
 
@@ -913,12 +934,14 @@ class Sale(
     def confirm(cls, sales):
         pool = Pool()
         Configuration = pool.get('sale.configuration')
+        transaction = Transaction()
+        context = transaction.context
         cls.set_sale_date(sales)
         cls.store_cache(sales)
         config = Configuration(1)
-        with Transaction().set_context(
-                queue_name='sale',
-                queue_scheduled_at=config.sale_process_after):
+        with transaction.set_context(
+                queue_scheduled_at=config.sale_process_after,
+                queue_batch=context.get('queue_batch', True)):
             cls.__queue__.process(sales)
 
     @classmethod
@@ -1070,8 +1093,7 @@ class SaleLine(TaxableMixin, sequence_ordered(), ModelSQL, ModelView):
         states={
             'readonly': ((Eval('sale_state') != 'draft')
                 & Bool(Eval('sale'))),
-            },
-        depends=['sale_state'])
+            })
     type = fields.Selection([
         ('line', 'Line'),
         ('subtotal', 'Subtotal'),
@@ -1080,22 +1102,19 @@ class SaleLine(TaxableMixin, sequence_ordered(), ModelSQL, ModelView):
         ], 'Type', select=True, required=True,
         states={
             'readonly': Eval('sale_state') != 'draft',
-            },
-        depends=['sale_state'])
+            })
     quantity = fields.Float(
         "Quantity", digits='unit',
         states={
             'invisible': Eval('type') != 'line',
             'required': Eval('type') == 'line',
             'readonly': Eval('sale_state') != 'draft',
-            },
-        depends=['type', 'sale_state'])
+            })
     actual_quantity = fields.Float(
         "Actual Quantity", digits='unit', readonly=True,
         states={
             'invisible': Eval('type') != 'line',
-            },
-        depends=['type'])
+            })
     unit = fields.Many2One('product.uom', 'Unit', ondelete='RESTRICT',
             states={
                 'required': Bool(Eval('product')),
@@ -1106,8 +1125,7 @@ class SaleLine(TaxableMixin, sequence_ordered(), ModelSQL, ModelView):
             If(Bool(Eval('product_uom_category')),
                 ('category', '=', Eval('product_uom_category')),
                 ('category', '!=', -1)),
-            ],
-        depends=['product', 'type', 'product_uom_category', 'sale_state'])
+            ])
     product = fields.Many2One('product.product', 'Product',
         ondelete='RESTRICT',
         domain=[
@@ -1134,7 +1152,7 @@ class SaleLine(TaxableMixin, sequence_ordered(), ModelSQL, ModelView):
             'taxes': Eval('taxes', []),
             'quantity': Eval('quantity'),
             },
-        depends=['type', 'sale_state', 'company'])
+        depends={'company'})
     product_uom_category = fields.Function(
         fields.Many2One('product.uom.category', 'Product Uom Category'),
         'on_change_with_product_uom_category')
@@ -1144,13 +1162,13 @@ class SaleLine(TaxableMixin, sequence_ordered(), ModelSQL, ModelView):
             'invisible': Eval('type') != 'line',
             'required': Eval('type') == 'line',
             'readonly': Eval('sale_state') != 'draft'
-            }, depends=['type', 'sale_state'])
+            })
     amount = fields.Function(Monetary(
             "Amount", digits='currency', currency='currency',
             states={
                 'invisible': ~Eval('type').in_(['line', 'subtotal']),
                 },
-            depends=['type', 'sale_state']), 'get_amount')
+            depends={'sale_state'}), 'get_amount')
     currency = fields.Function(
         fields.Many2One('currency.currency', 'Currency'),
         'on_change_with_currency')
@@ -1172,7 +1190,8 @@ class SaleLine(TaxableMixin, sequence_ordered(), ModelSQL, ModelView):
         states={
             'invisible': Eval('type') != 'line',
             'readonly': Eval('sale_state') != 'draft',
-            }, depends=['type', 'sale_state'])
+            },
+        depends={'sale'})
     invoice_lines = fields.One2Many('account.invoice.line', 'origin',
         'Invoice Lines', readonly=True)
     moves = fields.One2Many('stock.move', 'origin', 'Moves', readonly=True)
@@ -1192,8 +1211,7 @@ class SaleLine(TaxableMixin, sequence_ordered(), ModelSQL, ModelView):
     shipping_date = fields.Function(fields.Date('Shipping Date',
             states={
                 'invisible': Eval('type') != 'line',
-                },
-            depends=['type']),
+                }),
         'on_change_with_shipping_date')
     sale_state = fields.Function(
         fields.Selection('get_sale_states', "Sale State"),
@@ -1327,20 +1345,20 @@ class SaleLine(TaxableMixin, sequence_ordered(), ModelSQL, ModelView):
 
     @fields.depends('product', methods=['_get_tax_rule_pattern'])
     def compute_taxes(self, party):
-        taxes = set()
+        taxes = []
         pattern = self._get_tax_rule_pattern()
         for tax in self.product.customer_taxes_used:
             if party and party.customer_tax_rule:
                 tax_ids = party.customer_tax_rule.apply(tax, pattern)
                 if tax_ids:
-                    taxes.update(tax_ids)
+                    taxes.extend(tax_ids)
                 continue
-            taxes.add(tax.id)
+            taxes.append(tax.id)
         if party and party.customer_tax_rule:
             tax_ids = party.customer_tax_rule.apply(None, pattern)
             if tax_ids:
-                taxes.update(tax_ids)
-        return list(taxes)
+                taxes.extend(tax_ids)
+        return taxes
 
     @fields.depends('product', 'quantity', methods=['_get_context_sale_price'])
     def compute_unit_price(self):
@@ -1437,11 +1455,12 @@ class SaleLine(TaxableMixin, sequence_ordered(), ModelSQL, ModelView):
 
     @property
     @fields.depends(
-        'product', 'quantity', 'sale',
+        'product', 'quantity', 'company', 'sale',
         '_parent_sale.sale_date', '_parent_sale.shipping_date')
     def planned_shipping_date(self):
         pool = Pool()
         Date = pool.get('ir.date')
+        transaction = Transaction()
         if self.product and self.quantity is not None and self.quantity > 0:
             date = self.sale.sale_date if self.sale else None
             shipping_date = self.product.compute_shipping_date(date=date)
@@ -1450,7 +1469,12 @@ class SaleLine(TaxableMixin, sequence_ordered(), ModelSQL, ModelView):
             elif self.sale and self.sale.shipping_date:
                 shipping_date = max(shipping_date, self.sale.shipping_date)
             if shipping_date:
-                shipping_date = max(shipping_date, Date.today())
+                if self.company:
+                    company_id = self.company.id
+                else:
+                    company_id = transaction.context.get('company')
+                with transaction.set_context(company=company_id):
+                    shipping_date = max(shipping_date, Date.today())
             return shipping_date
 
     @fields.depends('sale', '_parent_sale.currency')
@@ -1681,23 +1705,18 @@ class SaleLine(TaxableMixin, sequence_ordered(), ModelSQL, ModelView):
             return
         moved_quantity = 0
         for move in self.moves:
-            if move.state != 'cancelled' and self.unit:
+            if move.state != 'cancelled':
                 moved_quantity += Uom.compute_qty(
-                    move.uom, move.quantity, self.unit, round=False)
+                    move.uom, move.quantity, self.unit)
         if self.quantity < 0:
             moved_quantity *= -1
         invoiced_quantity = 0
         for invoice_line in self.invoice_lines:
-            if (
-                    (not invoice_line.invoice
-                        or invoice_line.invoice.state != 'cancelled')
-                    and self.unit and invoice_line.unit):
+            if (not invoice_line.invoice
+                    or invoice_line.invoice.state != 'cancelled'):
                 invoiced_quantity += Uom.compute_qty(
-                    invoice_line.unit, invoice_line.quantity, self.unit,
-                    round=False)
+                    invoice_line.unit, invoice_line.quantity, self.unit)
         actual_quantity = max(moved_quantity, invoiced_quantity, key=abs)
-        if self.unit:
-            actual_quantity = self.unit.round(actual_quantity)
         if self.actual_quantity != actual_quantity:
             self.actual_quantity = actual_quantity
 
@@ -1801,7 +1820,8 @@ class SaleReport(CompanyReport):
         pool = Pool()
         Date = pool.get('ir.date')
         context = super().get_context(records, header, data)
-        context['today'] = Date.today()
+        with Transaction().set_context(company=header['company'].id):
+            context['today'] = Date.today()
         return context
 
 
@@ -1836,7 +1856,7 @@ class HandleShipmentExceptionAsk(ModelView):
     __name__ = 'sale.handle.shipment.exception.ask'
     recreate_moves = fields.Many2Many(
         'stock.move', None, None, 'Recreate Moves',
-        domain=[('id', 'in', Eval('domain_moves'))], depends=['domain_moves'])
+        domain=[('id', 'in', Eval('domain_moves'))])
     domain_moves = fields.Many2Many(
         'stock.move', None, None, 'Domain Moves')
 
@@ -1896,7 +1916,6 @@ class HandleInvoiceExceptionAsk(ModelView):
     recreate_invoices = fields.Many2Many(
         'account.invoice', None, None, 'Recreate Invoices',
         domain=[('id', 'in', Eval('domain_invoices'))],
-        depends=['domain_invoices'],
         help='The selected invoices will be recreated. '
         'The other ones will be ignored.')
     domain_invoices = fields.Many2Many(
